@@ -25,9 +25,12 @@ Schema dictionaries
 
 # Read:
 # * https://treyhunner.com/2019/04/why-you-shouldnt-inherit-from-list-and-dict-in-python/
+# * https://docs.python.org/3/library/abc.html
+# * https://docs.python.org/3/library/collections.abc.html
 
 from collections import OrderedDict
 from collections.abc import MutableMapping
+import re
 
 
 class SchemaError(Exception):
@@ -39,17 +42,25 @@ class Validators:
     """
     Collection of validator functions
 
-    All validator functions must accept three arguments in the order listed
+    All validator functions must accept four arguments in the order listed
     below. The actual variable names may differ depending on context.
 
     Args:
-        :value: value to be checked
+        :key: related dictionary key (used in error message)
+        :test_value: value to be checked
         :exp_value: expected value or comparison object
-        :key: dictionary key (used in error message)
+        :sd_instance: instance of the schemadict from which tests are called
+
+    Note:
+        * In most cases the last argument (the schemadict instance) is not
+          needed, and the test function may ignore the argument. However, in
+          case the schemadict is instantiated with custom validator functions,
+          these must be propagated when recursively checking nested data
+          structures
     """
 
-    @classmethod
-    def is_type(cls, value, exp_type, key):
+    @staticmethod
+    def is_type(key, value, exp_type, _):
         # Note: isinstance(True, int) evaluates to True
         if type(value) not in (exp_type,):
             raise TypeError(
@@ -58,7 +69,7 @@ class Validators:
             )
 
     @staticmethod
-    def is_gt(value, comp_value, key):
+    def is_gt(key, value, comp_value, _):
         if not value > comp_value:
             raise ValueError(
                 f"{key!r} too small: " +
@@ -66,15 +77,15 @@ class Validators:
             )
 
     @staticmethod
-    def is_lt(value, comp_value, key):
+    def is_lt(key, value, comp_value, _):
         if not value < comp_value:
             raise ValueError(
                 f"{key!r} too large: " +
                 f"expected < {comp_value!r}, but was {value!r}"
             )
 
-    @classmethod
-    def is_ge(cls, value, comp_value, key):
+    @staticmethod
+    def is_ge(key, value, comp_value, _):
         if not value >= comp_value:
             raise ValueError(
                 f"{key!r} too small: " +
@@ -82,7 +93,7 @@ class Validators:
             )
 
     @staticmethod
-    def is_le(value, comp_value, key):
+    def is_le(key, value, comp_value, _):
         if not value <= comp_value:
             raise ValueError(
                 f"{key!r} too large: " +
@@ -90,7 +101,7 @@ class Validators:
             )
 
     @staticmethod
-    def has_min_len(value, min_len, key):
+    def has_min_len(key, value, min_len, _):
         if not len(value) >= min_len:
             raise ValueError(
                 f"length of {key!r} too small: " +
@@ -98,7 +109,7 @@ class Validators:
             )
 
     @staticmethod
-    def has_max_len(value, max_len, key):
+    def has_max_len(key, value, max_len, _):
         if not len(value) <= max_len:
             raise ValueError(
                 f"length of {key!r} too large: " +
@@ -106,21 +117,36 @@ class Validators:
             )
 
     @staticmethod
-    def check_item_types(iterable, exp_item_type, key):
+    def check_regex_match(key, string, pattern, _):
+        if not re.match(pattern, string):
+            raise ValueError(
+                f"regex mismatch for {key!r}: " +
+                f"expected pattern {pattern!r}, got {string!r}"
+            )
+
+    @staticmethod
+    def check_item_types(key, iterable, exp_item_type, _):
         if not all(isinstance(item, exp_item_type) for item in iterable):
             raise TypeError(
                 f"unexpected type for item in iterable {key!r}: " +
                 f"expected {exp_item_type!r}"
             )
 
-    @classmethod
-    def check_item_schema(cls, iterable, item_schema, key):
+    @staticmethod
+    def check_item_schema(key, iterable, item_schema, sd_instance):
+        # TODO: check that iterables are not of type dict !?
         for item in iterable:
-            cls.check_schemadict(item, item_schema, key)
+            sd_instance._check_test_obj_against_test_funcs(key, item_schema, item)
+
+    @classmethod
+    def check_item_schemadict(cls, key, iterable, item_schema, sd_instance):
+        # TODO: check that iterables are of type dict !?
+        for item in iterable:
+            cls.check_schemadict(key, item, item_schema, sd_instance)
 
     @staticmethod
-    def check_schemadict(testdict, schema, key):
-        schemadict(schema).validate(testdict)
+    def check_schemadict(key, testdict, schema, sd_instance):
+        schemadict(schema, validators=sd_instance.validators).validate(testdict)
 
 
 class ValidatorDict(OrderedDict):
@@ -152,11 +178,19 @@ _VAL_COUNTABLE = {
     'max_len': Validators.has_max_len,
 }
 
+# Check string objects
+_VAL_STRING = {
+    **_VAL_COUNTABLE,
+    'regex': Validators.check_regex_match,
+}
+
 # Check iterable objects (list, tuple)
 _VAL_ITERABLE = {
     **_VAL_COUNTABLE,
     'item_types': Validators.check_item_types,
-    'item_schema': Validators.check_item_schema,
+    # TODO: find better names !!!!
+    'item_schema': Validators.check_item_schema,  # if each item is primitive type, but not a dictionary
+    'item_schemadict': Validators.check_item_schemadict,  # if each item is a complex dictionary
 }
 
 _VAL_SUBSCHEMA = {
@@ -169,7 +203,7 @@ STANDARD_VALIDATORS = ValidatorDict({
     bool: _VAL_TYPE,
     int: _VAL_NUM_REL,
     float: _VAL_NUM_REL,
-    str: _VAL_COUNTABLE,
+    str: _VAL_STRING,
     list: _VAL_ITERABLE,
     tuple: _VAL_ITERABLE,
     dict: _VAL_SUBSCHEMA,
@@ -181,11 +215,11 @@ METAKEY_CHECK_REQ_KEYS = '$required_keys'
 
 class schemadict(MutableMapping):
     """
-    A *schemadict* is a regular Python dictionary which specifies the type and
-    format of values for some given key. To check if a test dictionary is
-    conform with the expected schema, *schemadict* provides the `validate()`
-    method. If the test dictionary is ill-defined, an error will be thrown,
-    otherwise `None` is returned.
+    A *schemadict* is a dictionary that specifies the type and format of values
+    for some given key. To check if a test dictionary is conform with the
+    expected schema, *schemadict* provides the `validate()` method. If the test
+    dictionary is ill-defined, an error will be thrown, otherwise `None` is
+    returned.
     """
 
     def __init__(self, *args, validators=STANDARD_VALIDATORS, **kwargs):
@@ -265,18 +299,12 @@ class schemadict(MutableMapping):
         """
 
         # Check that testdict actually is a dictionary
-        Validators.is_type(testdict, dict, '$testdict')
-
-        # TODO: validate that the schema itself is valid
-
-        self._validate(testdict)
-
-    def _validate(self, testdict):
-        """Validate 'testdict' against 'self' (without schema validation)"""
+        Validators.is_type('$testdict', testdict, dict, self)
 
         for sd_key, sd_value in self.items():
-            # TODO: find better solution
+
             # Treat special keys/values separately
+            # TODO: find better solution
             if sd_key in self.metakeys:
                 self._check_special_key(sd_key, sd_value, testdict)
                 continue
@@ -287,7 +315,20 @@ class schemadict(MutableMapping):
             if td_value is None:
                 continue
 
-            for validator_key, validator_func in self.validators[sd_value['type']].items():
-                exp_value = sd_value.get(validator_key, None)
-                if exp_value is not None:
-                    validator_func(td_value, exp_value, sd_key)
+            self._check_test_obj_against_test_funcs(sd_key, sd_value, td_value)
+
+    def _check_test_obj_against_test_funcs(self, sd_key, sd_value, td_value):
+        """
+        For a given key, validate the test dictionary value against each test
+        function defined by the schemadict entry
+
+        Args:
+            :sd_key: common key for test dictionary and schemadict entry
+            :sd_value: mapping of test keywords and validator functions
+            :td_value: test dictionary value (object to test)
+        """
+
+        for validator_key, validator_func in self.validators[sd_value['type']].items():
+            exp_value = sd_value.get(validator_key, None)
+            if exp_value is not None:
+                validator_func(sd_key, td_value, exp_value, self)
